"""
========================================
Cluster-based active learning algorithms
========================================
"""

from sklearn.cluster.hierarchical import AgglomerativeClustering


class HierarchicalClustering:
    def __init__(self):
        return NotImplemented

    def __call__(self, *args, **kwargs):
        return NotImplemented

    def compute_errors(self):
        return NotImplemented
